import xarray as xr
import numpy as np
from typing import Union


class Compressor:
    def __init__(
        self,
        bits_per_pixel=10,
        mins=np.array(
            [
                -1.2278595,
                -2.5118103,
                -64.83977,
                63.404694,
                2.844452,
                199.10002,
                -17.254883,
                -26.29155,
                -1.1009827,
                -2.4184198,
                199.57048,
                198.95093,
            ]
        ),
        maxs=np.array(
            [
                103.90016,
                69.60857,
                339.15588,
                340.26526,
                317.86752,
                313.2767,
                315.99194,
                274.82297,
                93.786545,
                101.34922,
                249.91806,
                286.96323,
            ]
        ),
        variable_order=[
            "HRV",
            "IR_016",
            "IR_039",
            "IR_087",
            "IR_097",
            "IR_108",
            "IR_120",
            "IR_134",
            "VIS006",
            "VIS008",
            "WV_062",
            "WV_073",
        ],
    ):

        self.bits_per_pixel = bits_per_pixel
        self.mins = mins
        self.maxs = maxs
        self.variable_order = variable_order

    def fit(self, dataset: xr.Dataset, dims: list = ["time", "y", "x"]) -> None:
        """
        Calculate new min and max values for the compression

        Args:
            dataset: Xarray Dataset
            dims: Dims to compute over

        """
        self.mins = dataset.min(dims).compute()
        self.maxs = dataset.max(dims).compute()
        self.variable_order = dataset.coords["variable"].values

        print(f"The mins are: {self.mins}")
        print(f"The maxs are: {self.maxs}")
        print(f"The variable order is: {self.variable_order}")

    def compress(self, dataarray: xr.DataArray) -> Union[xr.DataArray, None]:
        """
        Compress Xarray DataArray to use 10-bit integers

        Args:
            dataarray: DataArray to compress

        Returns:
            The compressed DataArray
        """
        da_meta = dataarray.attrs

        for attr in ["mins", "maxs"]:
            assert (
                getattr(self, attr) is not None
            ), f"{attr} must be set in initialisation or through `fit`"

        dataarray = dataarray.reindex({"variable": self.variable_order}).transpose(
            "time", "y", "x", "variable"
        )

        upper_bound = (2 ** self.bits_per_pixel) - 1
        new_max = self.maxs - self.mins

        dataarray -= self.mins
        dataarray /= new_max
        dataarray *= upper_bound
        dataarray = dataarray.round().astype(np.int16)
        if is_dataset_clean(dataarray):
            dataarray.attrs = {"meta": str(da_meta)}  # Must be serializable

            return dataarray
        else:
            return None


def is_dataset_clean(dataarray: xr.DataArray) -> bool:
    """
    Checks if all the data values in a Dataset are not NaNs

    Args:
        dataarray: Xarray DataArray containing the data to check

    Returns:
        Bool of whether the dataset is clean or not
    """
    return (
        dataarray.notnull().compute().all().values and np.isfinite(dataarray).compute().all().values
    )



import datetime
from numbers import Number
from typing import Tuple, List

import numpy as np
import pyproj

# OSGB is also called "OSGB 1936 / British National Grid -- United
# Kingdom Ordnance Survey".  OSGB is used in many UK electricity
# system maps, and is used by the UK Met Office UKV model.  OSGB is a
# Transverse Mercator projection, using 'easting' and 'northing'
# coordinates which are in meters.  See https://epsg.io/27700
OSGB = 27700

# WGS84 is short for "World Geodetic System 1984", used in GPS. Uses
# latitude and longitude.
WGS84 = 4326
WGS84_CRS = f"EPSG:{WGS84}"

# Geographic bounds for various regions of interest, in order of min_lon, max_lon, min_lat, max_lat
GEOGRAPHIC_BOUNDS = {"UK": (-16, 45, 10, 62)}


class Transformers:
    """
    Class to store transformation from one Grid to another.

    Its good to make this only once, but need the
    option of updating them, due to out of data grids.
    """

    def __init__(self):
        """ Init """
        self._lat_lon_to_osgb = None
        self.make_transformers()

    def make_transformers(self):
        """
        Make transformers

         Nice to only make these once, as it makes calling the functions below quicker
        """
        self._lat_lon_to_osgb = pyproj.Transformer.from_crs(crs_from=WGS84, crs_to=OSGB)

    @property
    def lat_lon_to_osgb(self):
        """ lat-lon to OSGB property """
        return self._lat_lon_to_osgb


# make the transformers
transformers = Transformers()


def download_grids():
    """ The transformer grid sometimes need updating """
    pyproj.transformer.TransformerGroup(crs_from=WGS84, crs_to=OSGB).download_grids(verbose=True)

    transformers.make_transformers()


def lat_lon_to_osgb(lat: List[Number], lon: List[Number]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Change lat, lon to a OSGB coordinates

    Args:
        lat: latitude
        lon: longitude

    Return: 2-tuple of x (east-west), y (north-south).

    """
    return transformers.lat_lon_to_osgb.transform(lat, lon)



from satip.utils import (
    load_native_to_dataset,
    save_dataset_to_zarr,
    check_if_timestep_exists,
)
from satip.eumetsat import eumetsat_filename_to_datetime, eumetsat_cloud_name_to_datetime
import os
import pandas as pd
from pathlib import Path
import multiprocessing
from itertools import repeat
import xarray as xr
from tqdm import tqdm


def split_per_month(directory: str,
                    zarr_path: str,
                    hrv_zarr_path: str,
                    region: str,
                    temp_directory: str = "/mnt/ramdisk/",
                    spatial_chunk_size: int = 256,
                    temporal_chunk_size: int = 1, ):
    """
    Splits the Zarr creation into multiple, month-long Zarr files for parallel writing

    Args:
        directory: Top-level directory containing the compressed native files
        zarr_path: Path of the final Zarr file
        region: Name of the region to keep for the datastore
        spatial_chunk_size: Chunk size, in pixels in the x  and y directions, passed to Xarray
        temporal_chunk_size: Chunk size, in timesteps, for saving into the zarr file

    """

    # Get year
    temp_directory = Path(temp_directory)
    year_directories = os.listdir(directory)
    print(year_directories)
    dirs = []
    zarrs = []
    hrv_zarrs = []
    for year in year_directories:
        if not os.path.isdir(os.path.join(directory, year)):
            continue
        if year in ["2020", "2021"]:
            month_directories = os.listdir(os.path.join(directory, year))
            for month in month_directories:
                if not os.path.isdir(os.path.join(directory, year, month)):
                    continue
                month_directory = os.path.join(directory, year.split('/')[0], month.split('/')[0])
                month_zarr_path = zarr_path + f"_{year.split('/')[0]}_{month.split('/')[0]}.zarr"
                hrv_month_zarr_path = hrv_zarr_path + f"_{year.split('/')[0]}" \
                                                  f"_{month.split('/')[0]}.zarr"
                dirs.append(month_directory)
                zarrs.append(month_zarr_path)
                hrv_zarrs.append(hrv_month_zarr_path)
                zarr_exists = os.path.exists(month_zarr_path)
                if not zarr_exists:
                    # Inital zarr path before then appending
                    compressed_native_files = list(Path(month_directory).rglob("*.bz2"))
                    dataset, hrv_dataset = load_native_to_dataset(compressed_native_files[0],
                                                                  temp_directory,
                                                                  region)
                    save_dataset_to_zarr(dataset, zarr_path=month_zarr_path, zarr_mode="w")
                    save_dataset_to_zarr(hrv_dataset, zarr_path=hrv_month_zarr_path, zarr_mode="w")
    print(dirs)
    print(zarrs)
    pool = multiprocessing.Pool(processes=16)
    for _ in tqdm(pool.imap_unordered(
            wrapper,
            zip(
                dirs,
                zarrs,
                hrv_zarrs,
                repeat(temp_directory),
                repeat(region),
                repeat(spatial_chunk_size),
                repeat(temporal_chunk_size)
                ),
            )):
        print("Month done")


def wrapper(args):
    dirs, zarrs, hrv_zarrs, temp_directory, region, spatial_chunk_size, temporal_chunk_size = args
    create_or_update_zarr_with_native_files(dirs, zarrs, hrv_zarrs, temp_directory, region,
                                            spatial_chunk_size,
                                                temporal_chunk_size)


def create_or_update_zarr_with_native_files(
    directory: str,
    zarr_path: str,
    hrv_zarr_path: str,
        temp_directory: Path,
    region: str,
    spatial_chunk_size: int = 256,
    temporal_chunk_size: int = 1,
) -> None:
    """
    Creates or updates a zarr file with satellite native files

    Args:
        directory: Top-level directory containing the compressed native files
        zarr_path: Path of the final Zarr file
        region: Name of the region to keep for the datastore
        spatial_chunk_size: Chunk size, in pixels in the x  and y directions, passed to Xarray
        temporal_chunk_size: Chunk size, in timesteps, for saving into the zarr file

    """

    # Satpy Scene doesn't do well with fsspec
    compressed_native_files = list(Path(directory).rglob("*.bz2"))
    zarr_exists = os.path.exists(zarr_path)
    if zarr_exists:
        zarr_dataset = xr.open_zarr(zarr_path, consolidated=True)
        new_compressed_files = []
        for f in compressed_native_files:
            base_filename = f.name
            file_timestep = eumetsat_filename_to_datetime(str(base_filename))
            exists = check_if_timestep_exists(
                pd.Timestamp(file_timestep).round("5 min"), zarr_dataset
            )
            if not exists:
                new_compressed_files.append(f)
        compressed_native_files = new_compressed_files
    # Check if zarr already exists
    for entry in tqdm(compressed_native_files):
        try:
            dataset, hrv_dataset = load_native_to_dataset(entry, temp_directory, region)
            if dataset is not None and hrv_dataset is not None:
                try:
                    save_dataset_to_zarr(
                        dataset,
                        zarr_path=zarr_path,
                        x_size_per_chunk=spatial_chunk_size,
                        y_size_per_chunk=spatial_chunk_size,
                        timesteps_per_chunk=temporal_chunk_size,
                        channel_chunk_size=11
                    )
                    save_dataset_to_zarr(
                        hrv_dataset,
                        zarr_path=hrv_zarr_path,
                        x_size_per_chunk=spatial_chunk_size,
                        y_size_per_chunk=spatial_chunk_size,
                        timesteps_per_chunk=temporal_chunk_size,
                        channel_chunk_size=1
                    )
                except Exception as e:
                    print(f"Failed with: {e}")
            del dataset
            del hrv_dataset
        except Exception as e:
            print(f"Failed with Exception with {e}")


def pool_init(q):
    global processed_queue # make queue global in workers
    processed_queue = q

def native_wrapper(filename_and_area):
    filename, area = filename_and_area
    processed_queue.put(load_native_to_dataset(filename, area))
